@@ -3,7 +3,7 @@ from collections.abc import Iterable
 import psycopg
 from psycopg import sql, errors
 import pandas as pd
-from wrapg import util
+from wrapg import util, snippet
 
 
 # ===========================================================================
@@ -22,7 +22,7 @@ from wrapg import util
 # TODO: Add proper exceptions if parameters are missing or do not work
 # TODO: expose conn_import in init.py as class to easily modify attributes?? vs dict?
 conn_import: dict = {
-    "user": os.environ.get("PG_USERNAME"),
+    "user": os.environ.get("PG_USER"),
     "password": os.environ.get("PG_PASSWORD"),
     "host": os.environ.get("PG_HOST"),
     "dbname": os.environ.get("PG_DBNAME"),
@@ -178,39 +178,30 @@ def insert_ignore(
         # Open a cursor to perform database operations
         with conn.cursor() as cur:
 
+            # =================== Ignore_Insert Qry ==================
             # INSERT INTO table (name, email)
             # VALUES('Dave','dave@yahoo.com')
             # ON CONFLICT (name)
             # DO NOTHING;
 
-            # =================== Ignore_Insert Qry ==================
             try:
+                # uniform column names thru data submitted
                 if uniform == 1:
-                    qry = sql.SQL(
-                        "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO NOTHING"
-                    ).format(
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(map(sql.Identifier, columns)),
-                        sql.SQL(", ").join(map(sql.Placeholder, columns)),
-                        # this is technically known as 'conflict target'
-                        sql.SQL(", ").join(map(sql.Identifier, keys)),
+                    # get sql qry based on passed parameters
+                    qry = snippet.insert_ignore_snip(
+                        table=table, columns=columns, keys=keys
                     )
                     # print(qry.as_string(conn))
                     cur.executemany(query=qry, params_seq=rows)
 
                 else:
                     for row in rows:
-                        qry = sql.SQL(
-                            "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO NOTHING"
-                        ).format(
-                            sql.Identifier(table),
-                            sql.SQL(", ").join(map(sql.Identifier, tuple(row))),
-                            sql.SQL(", ").join(map(sql.Placeholder, tuple(row))),
-                            # this is technically known as 'conflict target'
-                            sql.SQL(", ").join(map(sql.Identifier, keys)),
+                        # get sql qry based on passed parameters for each row
+                        qry = snippet.insert_ignore_snip(
+                            table=table, columns=tuple(row), keys=keys
                         )
-                    # print(qry.as_string(conn))
-                    cur.execute(query=qry, params=row)
+                        # print(qry.as_string(conn))
+                        cur.execute(query=qry, params=row)
 
             # Catch no unique constriant error
             except errors.InvalidColumnReference as e:
@@ -218,35 +209,26 @@ def insert_ignore(
                 print("> Rolling back, attempt creation of new constriant...")
                 conn.rollback()
 
-                # Create new unique index & try again
                 try:
-                    uix_name = f'{table}_{"_".join(keys)}_uix'
-                    uix_sql = sql.SQL("CREATE UNIQUE INDEX {} ON {} ({});").format(
-                        sql.Identifier(uix_name),
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(map(sql.Identifier, keys)),
-                    )
-                    # print(idx_sql.as_string(conn))
+                    # Create new unique index & try insert_ignore again
+                    uix_sql = snippet.create_unique_index(table=table, keys=keys)
+                    # print(uix_sql.as_string(conn))
                     cur.execute(query=uix_sql)
 
-                    # Try again to execute qry
                     if uniform == 1:
+                        qry = snippet.insert_ignore_snip(
+                            table=table, columns=columns, keys=keys
+                        )
                         # Now execute previous insert_ignore statement
                         cur.executemany(query=qry, params_seq=rows)
 
                     else:
                         for row in rows:
-                            qry = sql.SQL(
-                                "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO NOTHING"
-                            ).format(
-                                sql.Identifier(table),
-                                sql.SQL(", ").join(map(sql.Identifier, tuple(row))),
-                                sql.SQL(", ").join(map(sql.Placeholder, tuple(row))),
-                                # this is technically known as 'conflict target'
-                                sql.SQL(", ").join(map(sql.Identifier, keys)),
+                            qry = snippet.insert_ignore_snip(
+                                table=table, columns=tuple(row), keys=keys
                             )
-                        # print(qry.as_string(conn))
-                        cur.execute(query=qry, params=row)
+                            # print(qry.as_string(conn))
+                            cur.execute(query=qry, params=row)
 
                 except Exception as indx_error:
                     print(">>> Error: ", indx_error)
@@ -300,67 +282,24 @@ def upsert(
         # Open a cursor to perform database operations
         with conn.cursor() as cur:
 
-            # Typical syntax
+            # =================== Upsert Qry ==================
             # INSERT INTO table (name, email)
             # VALUES('Dave','dave@yahoo.com')
             # ON CONFLICT (name)
             # DO UPDATE SET email=excluded.email
             # WHERE ...;
 
-            # Function to compose col=excluded.col sql for update
-            def set_str(cols: Iterable):
-                def exclude_sql(col):
-                    return sql.SQL("{}=EXCLUDED.{}").format(
-                        sql.Identifier(col),
-                        sql.Identifier(col),
-                    )
-
-                return map(exclude_sql, cols)
-
-            # =================== Upsert Qry ==================
-            # TODO: add WHERE clause to upsert qry; Do i need it?
-            """
-            WHERE excluded.validDate>phonebook2.validDate;
-            sudo code fro add where
-            def sql_to_format()
-                if where:
-                    return sql.SQL(statement with where)
-                else:
-                    return sql.SQL(statement without where)
-            """
-
-            # Silenced update_columns code as there is instances where
-            # you may need to update a key. timestamp when updating record
-            # leaving here for now in case i find issues later
-            # columns - keys, columns to update set etc
-            # update_columns = set(columns).difference(set(keys))
-
             try:
                 if uniform == 1:
-                    qry = sql.SQL(
-                        "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}"
-                    ).format(
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(map(sql.Identifier, columns)),
-                        sql.SQL(", ").join(map(sql.Placeholder, columns)),
-                        # this is technically known as 'conflict target'
-                        sql.SQL(", ").join(map(sql.Identifier, keys)),
-                        sql.SQL(", ").join(set_str(columns)),
-                    )
+                    qry = snippet.upsert_snip(table=table, columns=columns, keys=keys)
                     # print(qry.as_string(conn))
                     cur.executemany(query=qry, params_seq=rows)
 
                 else:
                     for row in rows:
-                        qry = sql.SQL(
-                            "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}"
-                        ).format(
-                            sql.Identifier(table),
-                            sql.SQL(", ").join(map(sql.Identifier, tuple(row))),
-                            sql.SQL(", ").join(map(sql.Placeholder, tuple(row))),
-                            # this is technically known as 'conflict target'
-                            sql.SQL(", ").join(map(sql.Identifier, keys)),
-                            sql.SQL(", ").join(set_str(tuple(row))),
+                        # Note tupe(row) returns column keys for each record
+                        qry = snippet.upsert_snip(
+                            table=table, columns=tuple(row), keys=keys
                         )
                         # print(qry.as_string(conn))
                         cur.execute(query=qry, params=row)
@@ -369,35 +308,27 @@ def upsert(
             except errors.InvalidColumnReference as e:
                 print(">>> Error: ", e)
                 print(f"> Creating unique index for {keys}...")
+                # !important, cannot attempt other operations after error unless rollback()
                 conn.rollback()
 
-                # Add unique index & try again
                 try:
-                    uix_name = f'{table}_{"_".join(keys)}_uix'
-                    uix_sql = sql.SQL("CREATE UNIQUE INDEX {} ON {} ({});").format(
-                        sql.Identifier(uix_name),
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(map(sql.Identifier, keys)),
-                    )
-                    # print(idx_sql.as_string(conn))
-
+                    # Create unique index & try upsert again
+                    uix_sql = snippet.create_unique_index(table=table, keys=keys)
+                    print(uix_sql.as_string(conn))
                     cur.execute(query=uix_sql)
 
                     if uniform == 1:
-                        # Now execute previous upsert statement
+                        qry = snippet.upsert_snip(
+                            table=table, columns=columns, keys=keys
+                        )
+                        # print(qry.as_string(conn))
                         cur.executemany(query=qry, params_seq=rows)
 
                     else:
                         for row in rows:
-                            qry = sql.SQL(
-                                "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}"
-                            ).format(
-                                sql.Identifier(table),
-                                sql.SQL(", ").join(map(sql.Identifier, tuple(row))),
-                                sql.SQL(", ").join(map(sql.Placeholder, tuple(row))),
-                                # this is technically known as 'conflict target'
-                                sql.SQL(", ").join(map(sql.Identifier, keys)),
-                                sql.SQL(", ").join(set_str(tuple(row))),
+                            # Note tupe(row) returns column keys for each record
+                            qry = snippet.upsert_snip(
+                                table=table, columns=tuple(row), keys=keys
                             )
                             # print(qry.as_string(conn))
                             cur.execute(query=qry, params=row)
@@ -432,7 +363,7 @@ def upsert_wo_idx():
     for novice users
 """
 
-# ================================= UPDATE ================================
+# ================================= UPDATE Function ================================
 def update(
     data: list[dict] | pd.DataFrame,
     table: str,
@@ -472,7 +403,7 @@ def update(
         # Open a cursor to perform database operations
         with conn.cursor() as cur:
 
-            # Syntax
+            # =================== Update Qry ===================
             # UPDATE table_name
             # SET column1 = value1,
             #     column2 = value2,
@@ -499,7 +430,6 @@ def update(
 
                 return map(set_sql, column_names)
 
-            # =================== Update Qry ==================
             if uniform == 1:
                 # print("> Uniform Data..")
                 qry = sql.SQL("UPDATE {} SET {} WHERE {}").format(
@@ -542,7 +472,6 @@ def create_table(table: str, columns: dict, conn_kwargs: dict = None):
     Returns:
         _type_: None
     """
-    # TODO: Add optional table constriants to function
 
     # Initialize conn_kwargs to empty dict if no arguments passed
     # Merge args into conn_final
@@ -558,6 +487,8 @@ def create_table(table: str, columns: dict, conn_kwargs: dict = None):
         # Open a cursor to perform database operations
         with conn.cursor() as cur:
 
+            # TODO: Add optional table constriants to function
+            # =================== Create Table Qry ===================
             # CREATE TABLE [IF NOT EXISTS] table_name (
             # column1 datatype(length) column_contraint,
             # column2 datatype(length) column_contraint,
@@ -576,14 +507,10 @@ def create_table(table: str, columns: dict, conn_kwargs: dict = None):
                 """
                 # function used to map to column names
                 def col_sql(col, value):
-                    return sql.SQL(f'"{col}" {value.upper()}')
-
-                    # TODO: Could not figure how to escape below? using .format()?
-                    # sql.SQL("{}").format(
-                    # sql.Identifier(col),
-                    # + sql.Literal(" ")
-                    # + sql.Literal(value.upper())
-                    # literal retruns quotes around value, not proper sql
+                    # return sql.SQL(f'"{col}" {value.upper()}')
+                    return sql.SQL("{} {}").format(
+                        sql.Identifier(col), sql.SQL(value.upper())
+                    )
 
                 return [col_sql(k, v) for k, v in column_info.items()]
 
@@ -638,8 +565,10 @@ def copy_from_csv(
         with conn.cursor() as cur:
 
             #! SPECIFIC COLUMNS NOT WORKING; Open ticket with pyscopg?
+            #=================== Copy Qry ===================
             # "COPY cust (name, age) FROM STDIN WITH (FORMAT csv)"
-            # Return composable sql statement
+            
+            # If csv has header
             if header:
                 copy_sql = sql.SQL(
                     "COPY {} FROM STDIN WITH (FORMAT csv, HEADER TRUE)"
@@ -658,7 +587,7 @@ def copy_from_csv(
 
                 with cur.copy(copy_sql) as copy:
 
-                    # using blocks/chunks
+                    # Using blocks/chunks
                     while data := f.read(block_size):
                         copy.write(data)
 
